// ----------------------------------------------------------------------------
//  :

class CAMERA {
public:

	double X0,Y0,Z0;	// 
    double **R;			//  
    double Phi,MaxPhi;	//    
	double dX0,dY0,dZ0;	//     (  )
    double Speed;		//   
	double StrafeSpeed;	//   
    bool SpeedRelative;	//       
    bool RotateSight;	//        
	bool RotateWith;	//       
    bool DrawAxes;		//  
    bool DrawPoint;		//  

    CAMERA();
    ~CAMERA();
    void RotateAround(double,double,double,double);		//   
};

// ----------------------------------------------------------------------------
// :

CAMERA::CAMERA()
{
   	X0 = 0;  Y0 = 0;  Z0 = 0;
    dX0 = 0;  dY0 = 0;  dZ0 = 0;
    MaxPhi = M_PI/6;
    Phi = M_PI/6;
    Speed = 1000;
    StrafeSpeed = 250;
    SpeedRelative = true;
    RotateSight = true;
    RotateWith = true;
    DrawAxes = true;
    DrawPoint = false;
    R = new double*[3];
    int i,j;
    for (i=0;i<3;i++) R[i] = new double[3];
    double R1[3][3] = {{1,  0,  0},
                       {0,  1,  0},
                       {0,  0,  1} };
    for(i=0;i<3;i++) for(j=0;j<3;j++) R[i][j] = R1[i][j];
};

// ----------------------------------------------------------------------------
// :

CAMERA::~CAMERA()
{
	for (int i=0; i<3; i++) delete[] R[i];
	delete[] R;
};

// ----------------------------------------------------------------------------
//   (dX0,dY0,dZ0)   W   Theta.   
//  ,    .

void CAMERA::RotateAround(double Wx, double Wy, double Wz, double Theta)
{
    //    (dX0,dY0,dZ0).
	//	:
	double W = sqrt(Wx*Wx + Wy*Wy + Wz*Wz);
    if (W < MINDOUBLE) return;
    Wx /= W;  Wy /= W;  Wz /= W;
    //      Z:
    double d = sqrt(Wx*Wx + Wy*Wy);
    if (d > MINDOUBLE) {
    	double sk = Wy/d;			double ck = Wx/d;
        double se = d;				double ce = Wz;
        double st = sin(Theta);		double ct = cos(Theta);
	    double R1[3][3] = {{ck*ce*ct + sk*st,  sk*ce*ct - ck*st,  -se*ct},
                           {ck*ce*st - sk*ct,  sk*ce*st + ck*ct,  -se*st},
                           {           ck*se,             sk*se,      ce} };
        double R2[3][3] = {{ck*ce,  -sk,  ck*se},
                           {sk*ce,   ck,  sk*se},
                           {  -se,    0,     ce} };
        double dX = R1[0][0]*dX0 + R1[0][1]*dY0 + R1[0][2]*dZ0;
        double dY = R1[1][0]*dX0 + R1[1][1]*dY0 + R1[1][2]*dZ0;
        double dZ = R1[2][0]*dX0 + R1[2][1]*dY0 + R1[2][2]*dZ0;
        dX0 = R2[0][0]*dX + R2[0][1]*dY + R2[0][2]*dZ;
        dY0 = R2[1][0]*dX + R2[1][1]*dY + R2[1][2]*dZ;
        dZ0 = R2[2][0]*dX + R2[2][1]*dY + R2[2][2]*dZ;
    }
    //	   Z:
    else {
        double st = sign(Wz)*sin(Theta);	double ct = cos(Theta);
        double R1[3][3] = {{ct, -st, 0},
                           {st,  ct, 0},
                           { 0,   0, 1} };
        double dX = R1[0][0]*dX0 + R1[0][1]*dY0 + R1[0][2]*dZ0;
        double dY = R1[1][0]*dX0 + R1[1][1]*dY0 + R1[1][2]*dZ0;
        double dZ = R1[2][0]*dX0 + R1[2][1]*dY0 + R1[2][2]*dZ0;
        dX0 = dX;  dY0 = dY;  dZ0 = dZ;
    }

    //	  .
    //	   .. :
    double WRx = R[0][0]*Wx + R[0][1]*Wy + R[0][2]*Wz;
    double WRy = R[1][0]*Wx + R[1][1]*Wy + R[1][2]*Wz;
    double WRz = R[2][0]*Wx + R[2][1]*Wy + R[2][2]*Wz;
	W = sqrt(WRx*WRx + WRy*WRy + WRz*WRz);
    if (W < MINDOUBLE) return;
    WRx /= W;  WRy /= W;  WRz /= W;
    //      Z:
    d = sqrt(WRx*WRx + WRy*WRy);
    if (d > MINDOUBLE) {
    	double sk = WRy/d;			double ck = WRx/d;
        double se = d;				double ce = WRz;
        double st = sin(-Theta);	double ct = cos(Theta);
	    double R1[3][3] = {{ck*ce*ct + sk*st,  sk*ce*ct - ck*st,  -se*ct},
                           {ck*ce*st - sk*ct,  sk*ce*st + ck*ct,  -se*st},
                           {           ck*se,             sk*se,      ce} };
        double R2[3][3] = {{ck*ce,  -sk,  ck*se},
                           {sk*ce,   ck,  sk*se},
                           {  -se,    0,     ce} };
        double R0[3][3];   int i,j,k;
        for (i=0; i<3; i++)
        	for (j=0; j<3; j++) {
            	R0[i][j] = 0;
            	for (k=0; k<3; k++) R0[i][j] += R2[i][k]*R1[k][j];
            }
        for (i=0; i<3; i++)
        	for (j=0; j<3; j++) {
            	R1[i][j] = 0;
            	for (k=0; k<3; k++) R1[i][j] += R0[i][k]*R[k][j];
            }
        for (i=0;i<3;i++)  for (j=0;j<3;j++)  R[i][j] = R1[i][j];
    }
    //	   Z:
    else {
        double st = sign(Wz)*sin(-Theta);	double ct = cos(Theta);
        double R0[3][3] = {{ct, -st, 0},
                           {st,  ct, 0},
                           { 0,   0, 1} };
        double R1[3][3];   int i,j,k;
        for (i=0; i<3; i++)
        	for (j=0; j<3; j++) {
            	R1[i][j] = 0;
            	for (k=0; k<3; k++) R1[i][j] += R0[i][k]*R[k][j];
            }
        for (i=0;i<3;i++)  for (j=0;j<3;j++)  R[i][j] = R1[i][j];
    }
};

// ----------------------------------------------------------------------------
//    :
/*
double **RotationMatrix(CAMERA *Camera)
{
/_
    //   :
    double cx = cos(Camera->Etta)*cos(Camera->Ksi);
    double cy = cos(Camera->Etta)*sin(Camera->Ksi);
    double cz = sin(Camera->Etta);
    //  :
    double d  = sqrt(cy*cy + cz*cz);
    double SD = sin(Camera->Delta);
    double CD = cos(Camera->Delta);
_/
	double s_k = sin(Camera->Ksi);    double c_k = cos(Camera->Ksi);
	double s_e = sin(Camera->Etta);   double c_e = cos(Camera->Etta);
    double c_kd = sin(-Camera->Ksi + Camera->Delta - M_PI/2);
    double s_kd = cos(-Camera->Ksi + Camera->Delta - M_PI/2);
//	double s_d = sin(Camera->Delta);
//    double c_d = cos(Camera->Delta);
    //  :
    double **R;
    int i,j;
    R = new double*[3];
    for (i=0;i<3;i++) R[i] = new double[3];
/_
    double d2 = (cy>=0) ? d : -d;
    double d3 = (cz>=0) ? 1 : -1;
    if (d > MINDOUBLE) {
	    double R1[3][3] = {{d3*d2*CD, -d3/d2*(cz*SD + cx*cy*CD),  d3/d2*(cy*SD - cx*cz*CD)},
	                       {d3*d2*SD,  d3/d2*(cz*CD - cx*cy*SD), -d3/d2*(cy*CD + cx*cz*SD)},
	                       {  cx,                     cy,                      cz} };
        for(i=0;i<3;i++) for(j=0;j<3;j++) R[i][j] = R1[i][j];
    }
    else {
	    double R2[3][3] = {{0, -SD, -CD},
	                       {0,  CD, -SD},
	                       {1,   0,   0} };
        for(i=0;i<3;i++) for(j=0;j<3;j++) R[i][j] = R2[i][j];
    }
_/
/_
	double R1[3][3] = {{c_k*c_e*c_d + s_k*s_d,  s_k*c_e*c_d - c_k*s_d,  -s_e*c_d},
                       {c_k*c_e*s_d - s_k*c_d,  s_k*c_e*s_d + c_k*c_d,  -s_e*s_d},
                       {              c_k*s_e,                s_k*s_e,       c_e} };
_/
	double R1[3][3] = {{-s_k*c_kd + c_k*c_e*s_kd,  c_k*c_kd + s_k*c_e*s_kd,  -s_e*s_kd},
                       {-s_k*s_kd - c_k*c_e*c_kd,  c_k*s_kd - s_k*c_e*c_kd,   s_e*c_kd},
                       {                 c_k*s_e,                  s_k*s_e,        c_e} };
    for(i=0;i<3;i++) for(j=0;j<3;j++) R[i][j] = R1[i][j];
    return R;
};
*/
